from shutil import which

from coalib.bearlib.abstractions.Linter import linter


@linter("bash", output_format="regex",
        output_regex=r'.+:(?P<line>.+):(?P<message>.*)')
class JavaPMDBear:
    """
    Check Java code for possible issues like potential bugs, dead code or too
    complicated expressions.

    More information is available at
    <http://pmd.github.io/pmd-5.4.1/pmd-java/rules/index.html>.
    """

    LANGUAGES = "Java"

    @classmethod
    def check_prerequisites(cls):  # pragma: no cover
        if which("bash") is None:
            return "bash is not installed."
        elif which("run.sh") is None:
            return ("PMD is missing. Make sure to install it from "
                    "<https://pmd.github.io/>")
        else:
            return True

    @staticmethod
    def create_arguments(filename, file, config_file,
                         check_best_practices: bool = True,
                         check_braces: bool = True,
                         check_clone_implementation: bool = True,
                         check_code_size: bool = True,
                         check_comments: bool = False,
                         check_controversial: bool = False,
                         check_design: bool = False,
                         check_imports: bool = True, check_naming: bool = True,
                         check_optimizations: bool = False,
                         check_strings: bool = False,
                         check_unnecessary: bool = True,
                         check_unused: bool = True):
        """
        :param check_best_practices:
            Activate to check for best practices.
        :param check_braces:
            Check for the right use of braces.
        :param check_clone_implementation:
            Check for the right implementation of the ``clone()`` function.
        :param check_code_size:
            Check for large or complicated code constructs.
        :param check_comments:
            Check comments for length, content and placement.
        :param check_controversial:
            Does various checks that are considered controversial.
        :param check_design:
            Check for optimal code implementations. Alternate approaches
            are suggested.
        :param check_imports:
            Check for duplicate and unused imports.
        :param check_naming:
            Check the names of identifiers against some rules.
        :param check_optimizations:
            Check for best pratices regarding optimization.
        :param check_strings:
            Check for String, StringBuffer and StringBuilder instances.
        :param check_unnecessary:
            Checks for unnecessary code.
        :param check_unused:
            Check for unused code.
        """
        options = {
            "java-basic": check_best_practices,
            "java-braces": check_braces,
            "java-clone": check_clone_implementation,
            "java-codesize": check_code_size,
            "java-comments": check_comments,
            "java-controversial": check_controversial,
            "java-design": check_design,
            "java-imports": check_imports,
            "java-naming": check_naming,
            "java-optimizations": check_optimizations,
            "java-strings": check_strings,
            "java-unnecessary": check_unnecessary,
            "java-unusedcode": check_unused}
        rules = ','.join(key for key in options if options[key])

        return which("run.sh"), "pmd", "-R", rules, "-d", filename



import re

from coalib.bearlib.abstractions.Lint import Lint
from coalib.bears.LocalBear import LocalBear


class GoTypeBear(LocalBear, Lint):
    executable = 'gotype'
    arguments = "-e {filename}"
    use_stderr = True
    output_regex = re.compile(
        r'(?P<file_name>.*):(?P<line>\d+):(?P<column>\d+):\s*(?P<message>.*)')
    LANGUAGES = "Go"

    def run(self, filename, file):
        '''
        Checks the code using ``gotype``. This will run gotype over each file
        seperately.
        '''
        return self.lint(filename)



from coalib.bearlib.abstractions.Linter import linter
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY


@linter(executable='scss-lint', output_format="regex",
        output_regex=r'.+:(?P<line>\d+)\s+(\[(?P<severity>.)\])\s*'
                     r'(?P<message>.*)',
        severity_map={'I': RESULT_SEVERITY.INFO,
                      'W': RESULT_SEVERITY.NORMAL,
                      'E': RESULT_SEVERITY.MAJOR})
class SCSSLintBear:
    """
    Check SCSS code to keep it clean and readable.

    More information is available at <https://github.com/brigade/scss-lint>.
    """

    LANGUAGES = "SCSS"

    @staticmethod
    def create_arguments(filename, file, config_file):
        return filename,



import json

from coalib.bearlib.abstractions.Linter import linter
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY
from coalib.results.Result import Result


@linter(executable='rubocop',
        use_stdin=True)
class RuboCopBear:
    """
    Check Ruby code for syntactic, formatting as well as semantic problems.

    See <https://github.com/bbatsov/rubocop#cops> for more information.
    """

    LANGUAGES = "Ruby"

    severity_map = {"error": RESULT_SEVERITY.MAJOR,
                    "warning": RESULT_SEVERITY.NORMAL,
                    "convention": RESULT_SEVERITY.INFO}

    @staticmethod
    def create_arguments(filename, file, config_file):
        # Need both stdin and filename. Explained in this comment:
        # https://github.com/bbatsov/rubocop/pull/2146#issuecomment-131403694
        return filename, '--stdin', '--format=json'

    def process_output(self, output, filename, file):
        output = json.loads(output)
        assert len(output['files']) == 1
        for result in output['files'][0]['offenses']:
            # TODO: Add condition for auto-correct, when rubocop is updated.
            # Relevant Issue: https://github.com/bbatsov/rubocop/issues/2932
            yield Result.from_values(
                origin="{class_name} ({rule})".format(
                    class_name=self.__class__.__name__,
                    rule=result['cop_name']),
                message=result['message'],
                file=filename,
                diffs=None,
                severity=self.severity_map[result['severity']],
                line=result['location']['line'],
                column=result['location']['column'],
                # Tested with linebreaks, it's secure.
                end_column=result['location']['column'] +
                result['location']['length'])



from coalib.bearlib.abstractions.Linter import linter


@linter(executable='gofmt',
        use_stdin=True,
        output_format='corrected',
        diff_message='Formatting can be improved.')
class GofmtBear:
    """
    Suggest better formatting options in Go code. Basic checks like alignment,
    indentation, and redundant parentheses are provided.

    This is done using the ``gofmt`` utility. For more information visit
    <https://golang.org/cmd/gofmt/>.
    """
    LANGUAGES = "Go"

    @staticmethod
    def create_arguments(filename, file, config_file):
        return ()



from shutil import which

from coalib.bearlib.abstractions.Linter import linter


@linter(executable="perl",
        output_format='regex',
        output_regex=r'.+:(?P<line>\d+):(?P<severity>.*) - (?P<message>.*)')
class VHDLLintBear:
    """
    Check VHDL code for common codestyle problems.

    Rules include:

     * Signals, variables, ports, types, subtypes, etc. must be lowercase.
     * Constants and generics must be uppercase.
     * Entities, architectures and packages must be "mixedcase" (may be 100%
       uppercase, but not 100% lowercase).
     * Ports must be suffixed using _i, _o or _io denoting its kind.
     * Labels must be placed in a separated line. Exception: component
       instantiation.
     * End statements must be documented indicating what are finishing.
     * Buffer ports are forbidden.
     * VHDL constructions of the "entity xxxx is" and similars must be in one
       line. You can't put "entity xxxxx" in one line and "is" in another.
     * No more than one VHDL construction is allowed in one line of code.

    See <http://fpgalibre.sourceforge.net/ingles.html#tp46> for more
    information.
    """

    LANGUAGES = "VHDL"

    @classmethod
    def check_prerequisites(cls):  # pragma: no cover
        if which("perl") is None:
            return "perl is not installed."
        elif which("bakalint.pl") is None:
            return ("bakalint is missing. Download it from "
                    "<http://fpgalibre.sourceforge.net/ingles.html#tp46> and "
                    "put it into your PATH.")
        else:
            return True

    @staticmethod
    def create_arguments(filename, file, config_file):
        return which("bakalint.pl"), "--input", filename



from coalib.bearlib.abstractions.Linter import linter


@linter(executable='cppclean',
        output_format='regex',
        output_regex=r'.+:(?P<line>\d+):(?P<message>.*)')
class CPPCleanBear:
    """
    Find problems in C++ source code that slow down development in large code
    bases. This includes finding unused code, among other features.

    Read more about available routines at
    <https://github.com/myint/cppclean#features>.
    """

    LANGUAGES = "C++"

    @staticmethod
    def create_arguments(filename, file, config_file):
        return filename,



import re

from coalib.bearlib.abstractions.Lint import Lint
from coalib.bears.LocalBear import LocalBear
from coalib.misc.Shell import escape_path_argument
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY


class PerlCriticBear(LocalBear, Lint):
    executable = 'perlcritic'
    output_regex = re.compile(
            r'(?P<line>\d+)\|(?P<column>\d+)\|(?P<severity>\d+)\|'
            r'(?P<origin>.*?)\|(?P<message>.*)')
    severity_map = {
        "1": RESULT_SEVERITY.MAJOR,
        "2": RESULT_SEVERITY.MAJOR,
        "3": RESULT_SEVERITY.NORMAL,
        "4": RESULT_SEVERITY.NORMAL,
        "5": RESULT_SEVERITY.INFO}
    LANGUAGES = "Perl"

    def run(self,
            filename,
            file,
            perlcritic_profile: str=""):
        '''
        Checks the code with perlcritic. This will run perlcritic over
        each of the files seperately

        :param perlcritic_profile: Location of the perlcriticrc config file.
        '''
        self.arguments = '--no-color --verbose "%l|%c|%s|%p|%m (%e)"'
        if perlcritic_profile:
            self.arguments += (" --profile "
                               + escape_path_argument(perlcritic_profile))
        self.arguments += " {filename}"
        return self.lint(filename)



import appdirs
import os
from shutil import copyfileobj
from urllib.request import urlopen

from bears import VERSION
from coalib.bearlib.abstractions.Linter import linter
from coalib.settings.Setting import path


DIR_PATH = os.path.dirname(os.path.abspath(__file__))

checkstyle_jar_file = os.path.join(DIR_PATH, 'checkstyle.jar')

known_checkstyles = {
    "google": "https://raw.githubusercontent.com/checkstyle/checkstyle/master/src/main/resources/google_checks.xml",
    "sun": 'https://raw.githubusercontent.com/checkstyle/checkstyle/master/src/main/resources/sun_checks.xml',
    "android-check-easy": "https://raw.githubusercontent.com/noveogroup/android-check/master/android-check-plugin/src/main/resources/checkstyle/checkstyle-easy.xml",
    "android-check-hard": "https://raw.githubusercontent.com/noveogroup/android-check/master/android-check-plugin/src/main/resources/checkstyle/checkstyle-hard.xml",
    "geosoft": "http://geosoft.no/development/geosoft_checks.xml"}


def known_checkstyle_or_path(setting):
    if str(setting) in known_checkstyles.keys():
        return str(setting)
    else:
        return path(setting)


@linter(executable='java',
        output_format='regex',
        output_regex=r'\[(?P<severity>WARN|INFO)\] *.+:'
                     r'(?P<line>\d+)(?::(?P<column>\d+))?: *'
                     r'(?P<message>.*?) *\[(?P<origin>[a-zA-Z]+?)\]',
        prerequisite_check_command=('java', '-jar', checkstyle_jar_file, '-v'),
        prerequisite_check_fail_message='jar file ' + checkstyle_jar_file +
                                        ' not found.')
class CheckstyleBear:
    """
    Check Java code for possible style, semantic and design issues.

    For more information, consult
    <http://checkstyle.sourceforge.net/checks.html>.
    """

    LANGUAGES = "Java"

    def create_arguments(
            self, filename, file, config_file,
            checkstyle_configs: known_checkstyle_or_path="google"):
        """
        :param checkstyle_configs:
            A file containing configs to use in ``checkstyle``. It can also
            have the special values:

            - google - Google's Java style. More info at
              <http://checkstyle.sourceforge.net/style_configs.html>.
            - sun - Sun's Java style. These are the same
              as the default Eclipse checks. More info at
              <http://checkstyle.sourceforge.net/style_configs.html>.
            - android-check-easy - The easy Android configs provided by the
              android-check eclipse plugin. More info at
              <https://github.com/noveogroup/android-check>.
            - android-check-hard - The hard Android confis provided by the
              android-check eclipse plugin. More info at
              <https://github.com/noveogroup/android-check>.
            - geosoft - The Java style followed by GeoSoft. More info at
              <http://geosoft.no/development/javastyle.html>
        """
        checkstyle_configs = self.known_configs(checkstyle_configs)

        return ('-jar', checkstyle_jar_file, '-c', checkstyle_configs, filename)

    def known_configs(self, key):
        url = known_checkstyles.get(key, None)
        if url is None:  # If not known, it must be custom path
            return key

        filename = os.path.join(self.data_dir, key + "_style.xml")
        if os.path.exists(filename):
            return filename

        self.info("Downloading checkstyle configs for {key} from {url}"
                  .format(key=key, url=url))
        with urlopen(url) as response, open(filename, 'wb') as out_file:
            copyfileobj(response, out_file)
        return filename

    @property
    def data_dir(self):
        _data_dir = os.path.abspath(os.path.join(
            appdirs.user_data_dir('coala-bears', version=VERSION),
            self.name))

        if not os.path.isdir(_data_dir):  # pragma: no cover
            os.makedirs(_data_dir)
        return _data_dir



from shutil import which
import xml.etree.ElementTree as ET

from coalib.bears.GlobalBear import GlobalBear
from coalib.misc.Shell import run_shell_command
from coalib.results.Result import Result
from coalib.results.SourceRange import SourceRange


class CPDBear(GlobalBear):

    LANGUAGES = ('C++', 'C#', 'Objective C', 'Java', 'JavaScript', 'Fortran',
                 'Go', 'JSP', 'Python 2', 'Python 3', 'Python', 'Ruby',
                 'PHP', 'Scala')

    @classmethod
    def check_prerequisites(cls):  # pragma: no cover
        if which("bash") is None:
            return "bash is not installed."
        elif which("run.sh") is None:
            return ("CPD is missing. Make sure to install it from "
                    "<https://pmd.github.io/>")
        else:
            return True

    def run(self, language: str, minimum_tokens: int=10,
            ignore_annotations: bool=False, ignore_identifiers: bool=True,
            ignore_literals: bool=False, ignore_usings: bool=False,
            skip_duplicate_files: bool=True):
        """
        Checks for similar code that looks as it could be replaced to reduce
        redundancy.

        For more details see:
        <https://pmd.github.io/pmd-5.4.1/usage/cpd-usage.html>

        :param language:
            One of "cpp", "cs", "js", "f", "go", "jsp", "m", "php", "py",
            "rb", "scala", "java" (i.e. the file ending of your language).
        :param minimum_tokens:
            The minimum token length which should be reported as a duplicate.
        :param ignore_annotations:
            Ignore language annotations when comparing text.
        :param ignore_identifiers:
            Ignore constant and variable names when comparing text.
        :param ignore_literals:
            Ignore number values and string contents when comparing text.
        :param ignore_usings:
            Ignore ``using`` directives in C#.
        :param skip_duplicate_files:
            Ignore multiple copies of files of the same name and length in
            omparison.
        """
        language_args = {
            "cpp": "cpp",
            "cs": "cs",
            "js": "ecmascript",
            "f": "fortran",
            "go": "go",
            "jsp": "jsp",
            "m": "objectivec",
            "php": "php",
            "py": "python",
            "rb": "ruby",
            "scala": "scala",
            "java": "java"}

        if language not in language_args:  # pragma: no cover
            self.err("This bear does not support files with the extension "
                     "'{}'.".format(language))
            return

        options = {
            "--ignore-annotations": ignore_annotations,
            "--ignore-identifiers": ignore_identifiers,
            "--ignore-literals": ignore_literals,
            "--ignore-usings": ignore_usings,
            "--skip-duplicate-files": skip_duplicate_files}

        files = ",".join(self.file_dict.keys())
        arguments = ("bash", which("run.sh"), "cpd", "--skip-lexical-errors",
                     "--minimum-tokens", str(minimum_tokens),
                     "--language", language_args[language],
                     "--files", files,
                     "--format", "xml")

        arguments += tuple(option
                           for option, enable in options.items()
                           if enable is True)

        stdout_output, _ = run_shell_command(arguments)

        if stdout_output:
            root = ET.fromstring(stdout_output)

            for duplication in root.findall('duplication'):
                length = int(duplication.attrib['lines'])
                affected_code = list()

                for file in duplication.findall('file'):
                    filename = file.attrib['path']
                    start_line = int(file.attrib['line'])
                    end_line = start_line + length - 1

                    affected_code.append(
                        SourceRange.from_values(filename,
                                                start_line=start_line,
                                                end_line=end_line))

                yield Result(
                    self, "Duplicate code found.", affected_code,
                    additional_info=(
                        "Duplicate code is an indicator "
                        "that you have more code than you need. Consider"
                        " refactor your code to remove one of the"
                        " occurrences. For more information go here:"
                        "http://tinyurl.com/coala-clone"))



from coalib.bearlib.abstractions.Linter import linter


@linter(executable='yamllint',
        output_format="regex",
        output_regex=r'.+:(?P<line>\d+):(?P<column>\d+): '
                     r'\[(?P<severity>error|warning)\] (?P<message>.+)')
class YAMLLintBear:
    """
    Check yaml code for errors and possible problems.

    You can read more about capabilities at
    <http://yamllint.readthedocs.org/en/latest/rules.html>.
    """

    LANGUAGES = "YAML"

    @staticmethod
    def create_arguments(filename, file, config_file, yamllint_config: str=''):
        """
        :param yamllint_config: Path to a custom configuration file.
        """
        args = ('-f', 'parsable', filename)
        if yamllint_config:
            args += ('--config=' + yamllint_config,)
        return args



import json

from coalib.bearlib.abstractions.Linter import linter
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY
from coalib.results.Result import Result


@linter(executable='dockerfile_lint')
class DockerfileLintBear:
    """
    Check file syntax as well as arbitrary semantic and best practice
    in Dockerfiles. it also checks LABEL rules against docker images.

    Uses ``dockerfile_lint`` to provide the analysis.
    See <https://github.com/projectatomic/dockerfile_lint#dockerfile-lint> for
    more information .
    """
    LANGUAGES = "Dockerfile"

    severity_map = {
        "error": RESULT_SEVERITY.MAJOR,
        "warn": RESULT_SEVERITY.NORMAL,
        "info": RESULT_SEVERITY.INFO}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return '--json', '-f', filename

    def process_output(self, output, filename, file):
        output = json.loads(output)

        for severity in output:
            if severity == "summary":
                continue
            for issue in output[severity]["data"]:
                yield Result.from_values(
                    origin=self,
                    message=issue["message"],
                    file=filename,
                    severity=self.severity_map[issue["level"]],
                    line=issue["line"])



import os
import sys


VERSION_FILE = os.path.join(os.path.dirname(__file__), "VERSION")
with open(VERSION_FILE, 'r') as ver:
    VERSION = ver.readline().strip()


__version__ = VERSION


def assert_supported_version():  # pragma: no cover
    if not sys.version_info > (3, 2):
        print("coala supports only python 3.3 or later.")
        exit(4)



import platform
from shlex import split

from coalib.bearlib.abstractions.Linter import linter
from coalib.bearlib.spacing.SpacingHelper import SpacingHelper


@linter(executable="indent" if platform.system() != "Darwin" else "gindent",
        use_stdin=True,
        output_format='corrected',
        diff_message="Indentation can be improved.")
class GNUIndentBear:
    """
    This bear checks and corrects spacing and indentation via the well known
    Indent utility.

    C++ support is considered experimental.
    """

    LANGUAGES = ("C", "C++")

    @staticmethod
    def create_arguments(filename, file, config_file,
                         max_line_length: int=80,
                         use_spaces: bool=True,
                         tab_width: int=SpacingHelper.DEFAULT_TAB_WIDTH,
                         indent_cli_options: str=''):
        """
        :param max_line_length:    Maximum number of characters for a line.
        :param use_spaces:         True if spaces are to be used, else tabs.
        :param tab_width:          Number of spaces per indent level.
        :param indent_cli_options: Any command line options the indent binary
                                   understands. They will be simply passed
                                   through.
        """
        return ("--no-tabs" if use_spaces else "--use-tabs",
                "--line-length", str(max_line_length),
                "--indent-level", str(tab_width),
                "--tab-size", str(tab_width)) + tuple(split(indent_cli_options))



from os.path import abspath, dirname, join

from coalib.bearlib.abstractions.Linter import linter


jar = join(dirname(abspath(__file__)), 'scalastyle.jar')
scalastyle_config_file = join(dirname(abspath(__file__)),
                              'scalastyle_config.xml')


@linter(executable='java',
        output_format='regex',
        output_regex=r'(?P<severity>warning) file=.+ message=(?P<message>.+) '
                     r'line=(?P<line>\d+)(?: column=(?P<column>\d+))?',
        prerequisite_check_command=('java', '-jar', jar, '-c',
                                    scalastyle_config_file, '.', '-q', 'true'),
        prerequisite_check_fail_message='Required jar file ' + jar +
                                        ' not found.')
class ScalaLintBear:
    """
    Check Scala code for codestyle, but also semantical problems,
    e.g. cyclomatic complexity.
    """

    LANGUAGES = "Scala"

    @staticmethod
    def create_arguments(filename, file, config_file,
                         scalalint_config: str=scalastyle_config_file):
        """
        :param scalalint_config: Path to a custom configuration file.
        """
        return '-jar', jar, filename, '--config', scalalint_config
